@@ -36,6 +36,7 @@ except ImportError:
     os.environ["PATH"] += os.pathsep + ncscliPath
     import ncs
 import jsonToKnownHosts
+import ncsgeth
 import tellInstances
 
 logger = logging.getLogger(__name__)
@@ -136,6 +137,18 @@ def ingestJson( srcFilePath, dbName, collectionName, append ):
     #logger.info( 'cmd: %s', cmd )
     subprocess.check_call( cmd )
 
+def importAnchorNodes( srcFilePath, db, collectionName ):
+    # uses some args from the global ArgumentParser
+    cmd = [
+        'mongoimport', '--host', args.mongoHost, '--port', str(args.mongoPort),
+        '-d', db.name, '-c', collectionName, '--jsonArray',
+        '--mode=upsert', '--upsertFields=instanceId',
+        srcFilePath
+    ]
+    #logger.info( 'cmd: %s', cmd )
+    subprocess.check_call( cmd )
+    db[collectionName].create_index( 'instanceId' )
+
 def launchInstances( authToken, nInstances, sshClientKeyName, launchedJsonFilepath,
         filtersJson=None, encryptFiles=True ):
     returnCode = 13
@@ -156,7 +169,7 @@ def terminateNcsScInstances( authToken, instanceIds ):
     dateTimeStr = datetime.datetime.now( datetime.timezone.utc ).isoformat()
     try:
         ncs.terminateInstances( authToken, instanceIds )
-        logger.info( 'terminateInstances returned' )
+        logger.debug( 'terminateInstances returned' )
     except Exception as exc:
         logger.warning( 'got exception terminating %d instances (%s) %s', 
             len( instanceIds ), type(exc), exc )
@@ -245,7 +258,7 @@ def recruitInstances( nWorkersWanted, launchedJsonFilePath, launchWanted,
         stepStatuses = tellInstances.tellInstances( startedInstances, installerCmd,
             resultsLogFilePath=resultsLogFilePath,
             download=None, downloadDestDir=None, jsonOut=None, sshAgent=args.sshAgent,
-            timeLimit=args.timeLimit, upload=args.uploads, stopOnSigterm=False,
+            timeLimit=args.timeLimit, upload=args.uploads, stopOnSigterm=True,
             knownHostsOnly=False
             )
         # COULD restore our handler because tellInstances may have overridden it
@@ -339,7 +352,7 @@ def launchEdgeNodes( dataDirPath, db, args ):
         mQuery =  {'state': 'checked' }
         nExisting = db['checkedInstances'].count_documents( mQuery )
         logger.info( '%d workers checked', nExisting )
-        nToLaunch = int( (args.target - nExisting) * 1.5 )
+        nToLaunch = int( (args.target - nExisting) * 3 )
         nToLaunch = max( nToLaunch, 0 )
 
         nAvail = ncs.getAvailableDeviceCount( args.authToken, filtersJson=args.filter )
@@ -392,11 +405,11 @@ def checkGethProcesses( instances, dataDirPath ):
     # check for a running geth process on each instance
     stepStatuses = tellInstances.tellInstances( instances, cmd,
         timeLimit=15*60,
-        resultsLogFilePath = dataDirPath + '/checkProcesses.jlog',
-        knownHostsOnly=True
+        resultsLogFilePath = dataDirPath + '/checkGethProcesses.jlog',
+        knownHostsOnly=True, stopOnSigterm=True
         )
     #logger.info( 'proc statuses: %s', stepStatuses )
-    errorsByIid = {status['instanceId']: status['status'] for status in stepStatuses if status['status'] }
+    errorsByIid = {status['instanceId']: status for status in stepStatuses if status['status'] }
     logger.info( 'errorsByIid: %s', errorsByIid )
     return errorsByIid
 
@@ -409,7 +422,7 @@ def checkInstanceClocks( liveInstances, dataDirPath ):
     stepStatuses = tellInstances.tellInstances( liveInstances, cmd,
         timeLimit=2*60,
         resultsLogFilePath = jlogFilePath,
-        knownHostsOnly=True, sshAgent=not True
+        knownHostsOnly=True, sshAgent=not True, stopOnSigterm=True
         )
     #logger.info( 'proc statuses: %s', stepStatuses )
     errorsByIid = {status['instanceId']: status for status in stepStatuses if status['status'] }
@@ -445,6 +458,190 @@ def checkInstanceClocks( liveInstances, dataDirPath ):
     logger.info( '%d errorsByIid: %s', len(errorsByIid), errorsByIid )
     return errorsByIid
 
+def retrieveLogs( goodInstances, dataDirPath, farmDirPath ):
+    '''retrieve logs from nodes, storing them in dataDirPath/nodeLogs subdirectories'''
+    logger.info( 'retrieving logs for %d instance(s)', len(goodInstances) )
+
+    nodeLogFilePath = farmDirPath + '/geth.log'
+
+    # download the log file from each instance
+    stepStatuses = tellInstances.tellInstances( goodInstances,
+        download=nodeLogFilePath, downloadDestDir=dataDirPath +'/nodeLogs',
+        timeLimit=15*60,
+        knownHostsOnly=True, stopOnSigterm=True
+        )
+    #logger.info( 'download statuses: %s', stepStatuses )
+    errorsByIid = {status['instanceId']: status['status'] for status in stepStatuses if status['status'] }
+    return errorsByIid
+
+def checkLogs( liveInstances, dataDirPath ):
+    '''check contents of logs from nodes, save errorSummary.csv, return errorsByIid'''
+    logsDirPath = os.path.join( dataDirPath, 'nodeLogs' )
+    logDirs = os.listdir( logsDirPath )
+    logger.info( '%d logDirs found', len(logDirs ) )
+
+    liveIids = [inst['instanceId'] for inst in liveInstances]
+
+    errorsByIid = {}
+
+    for logDir in logDirs:
+        iid = logDir
+        iidAbbrev = iid[0:8]
+        if iid not in liveIids:
+            logger.debug( 'not checking non-live instance %s', iidAbbrev )
+            continue
+        logFilePath = os.path.join( logsDirPath, logDir, 'geth.log' )
+        if not os.path.isfile( logFilePath ):
+            logger.warning( 'no log file %s', logFilePath )
+        elif os.path.getsize( logFilePath ) <= 0:
+            logger.warning( 'empty log file "%s"', logFilePath )
+        else:
+            connected = False
+            sealed = False
+            latestSealLine = None
+            with open( logFilePath ) as logFile:
+                for line in logFile:
+                    line = line.rstrip()
+                    if not line:
+                        continue
+                    if 'ERROR' in line:
+                        if 'Ethereum peer removal failed' in line:
+                            logger.debug( 'ignoring "peer removal failed" for %s', iidAbbrev )
+                            continue
+                        logger.info( '%s: %s', iidAbbrev, line )
+                        theseErrors = errorsByIid.get( iid, [] )
+                        theseErrors.append( line )
+                        errorsByIid[iid] = theseErrors
+                    elif 'Started P2P networking' in line:
+                        connected = True
+                    elif 'Successfully sealed' in line:
+                        sealed = True
+                        latestSealLine = line
+                if not connected:
+                    logger.warning( 'instance %s did not initialize fully', iidAbbrev )
+                if sealed:
+                    #logger.info( 'instance %s has successfully sealed', iidAbbrev )
+                    logger.debug( '%s latestSealLine: %s', iidAbbrev, latestSealLine )
+                    pat = r'\[(.*\|.*)\]'
+                    dateStr = re.search( pat, latestSealLine ).group(1)
+                    if dateStr:
+                        dateStr = dateStr.replace( '|', ' ' )
+                        latestSealDateTime = dateutil.parser.parse( dateStr )
+                        logger.info( 'latestSealDateTime: %s', latestSealDateTime.isoformat() )
+
+    logger.info( 'found errors for %d instance(s)', len(errorsByIid) )
+    #print( 'errorsByIid', errorsByIid  )
+    summaryCsvFilePath = os.path.join( dataDirPath, 'errorSummary.csv' )
+    fileExisted = os.path.isfile( summaryCsvFilePath )
+    with open( summaryCsvFilePath, 'a' ) as csvOutFile:
+        if not fileExisted:
+            print( 'instanceId', 'msg',
+                sep=',', file=csvOutFile
+                )
+        for iid in errorsByIid:
+            lines = errorsByIid[iid]
+            logger.info( 'saving %d errors for %s', len(lines), iid[0:8])
+            for line in lines:
+                print( iid, line,
+                    sep=',', file=csvOutFile
+                )
+    return errorsByIid
+
+def lastGenDateTime( coll ):
+    '''return modification dateTime of a mongo collection, as a (hopefully) tz-aware datetime'''
+    found = coll.find().sort([( '$natural', -1 )]).limit(1)
+    if found:
+        return found[0]['_id'].generation_time
+    return None
+
+def ingestGethLog( logFile, coll ):
+    '''ingest records from a geth node log into a mongo collection'''
+    logger.debug( 'would ingest %s into %s', logFile.name, coll.name )
+    datePat = r'\[(.*\|.*)\]'
+    lineDateTime = datetime.datetime.fromtimestamp( 0, datetime.timezone.utc )
+    lineLevel = None
+    recs = []
+    for line in logFile:
+        line = line.rstrip()
+        if not line:
+            continue
+        levelPart = line.split()[0]
+        if levelPart in ['INFO', 'WARN', 'ERROR']:
+            lineLevel = levelPart
+
+        lineDateTime = None
+        dateStr = re.search( datePat, line ).group(1)
+        if dateStr:
+            dateStr = dateStr.replace( '|', ' ' )
+            lineDateTime = dateutil.parser.parse( dateStr )
+        msg = line
+        recs.append( {
+            'dateTime': lineDateTime.isoformat(),
+            'level': lineLevel,
+            'msg': msg,
+            } )
+    if len(recs):
+        coll.insert_many( recs, ordered=True )
+        coll.create_index( 'dateTime' )
+        coll.create_index( 'level' )
+
+def processNodeLogFiles( dataDirPath ):
+    '''ingest all new or updated node logs; delete very old log files'''
+    logsDirPath = os.path.join( dataDirPath, 'nodeLogs' )
+    logDirs = os.listdir( logsDirPath )
+    logger.info( '%d logDirs found', len(logDirs ) )
+    # but first, delete very old log files
+    lookbackDays = 7
+    thresholdDateTime = datetime.datetime.now( datetime.timezone.utc ) \
+        - datetime.timedelta( days=lookbackDays )
+    #lookBackHours = 24
+    #newerThresholdDateTime = datetime.datetime.now( datetime.timezone.utc ) \
+    #    - datetime.timedelta( hours=lookBackHours )
+
+    for logDir in logDirs:
+        inFilePath = os.path.join( logsDirPath, logDir, 'geth.log' )
+        if os.path.isfile( inFilePath ):
+            fileModDateTime = datetime.datetime.fromtimestamp( os.path.getmtime( inFilePath ) )
+            fileModDateTime = universalizeDateTime( fileModDateTime )
+            if fileModDateTime <= thresholdDateTime:
+                os.remove( inFilePath )
+            #elif fileModDateTime <= newerThresholdDateTime:
+            #    logger.warning( 'old fileModDateTime for %s (%s)', logDir, fileModDateTime )
+    
+    # ingest all new or updated node logs
+    loggedCollNames = db.list_collection_names(
+        filter={ 'name': {'$regex': r'^nodeLog_.*'} } )
+    for logDir in logDirs:
+        # logDir is also the instanceId
+        inFilePath = os.path.join( logsDirPath, logDir, 'geth.log' )
+        if not os.path.isfile( inFilePath ):
+            logger.info( 'no file "%s"', inFilePath )
+            continue
+        elif os.path.getsize( inFilePath ) <= 0:
+            logger.warning( 'empty log file "%s"', inFilePath )
+            continue
+        collName = 'nodeLog_' + logDir
+        if collName in loggedCollNames:
+            existingGenTime = lastGenDateTime( db[collName] ) - datetime.timedelta(hours=1)
+            fileModDateTime = datetime.datetime.fromtimestamp( os.path.getmtime( inFilePath ) )
+            fileModDateTime = universalizeDateTime( fileModDateTime )
+            if existingGenTime >= fileModDateTime:
+                #logger.info( 'already posted %s %s %s',
+                #    logDir[0:8], fmtDt( existingGenTime ), fmtDt( fileModDateTime )  )
+                continue
+        logger.info( 'ingesting log for %s', logDir[0:16] )
+        try:
+            with open( inFilePath, 'r' ) as logFile:
+                # for safety, ingest to a temp collection and then rename (with replace) when done
+                ingestGethLog( logFile, db['nodeLog_temp'] )
+                try:
+                    db['nodeLog_temp'].rename( collName, dropTarget=True )
+                except Exception as exc:
+                    logger.warning( 'exception (%s) renaming temp collection %s', type(exc), exc, exc_info=False )
+                #logger.debug( 'ingested %s', collName )
+        except Exception as exc:
+            logger.warning( 'exception (%s) ingesting %s', type(exc), inFilePath, exc_info=False )
+
 def checkEdgeNodes( dataDirPath, db, args ):
     '''checks status of edge nodes, updates database'''
     if not db.list_collection_names():
@@ -518,6 +715,7 @@ def checkEdgeNodes( dataDirPath, db, args ):
             upsert=True
             )
     checkables = [inst for inst in checkables if inst['instanceId'] in liveIidSet]
+    logger.info( '%d instances checkable', len(checkables) )
 
     errorsByIid = checkInstanceClocks( liveInstances, dataDirPath )
     for iid, err in errorsByIid.items():
@@ -528,35 +726,80 @@ def checkEdgeNodes( dataDirPath, db, args ):
             coll.update_one( {'_id': iid}, { "$inc": { "nFailures": 1 } } )
         elif 'status' in err:
             status = err['status']
-            logger.warning( 'instance %s gave status (%s) "%s"', iid[0:8], type(status), status )
             if isinstance( status, Exception ):
                 coll.update_one( {'_id': iid}, { "$inc": { "nExceptions": 1 } } )
             else:
                 coll.update_one( {'_id': iid}, { "$inc": { "nFailures": 1 } } )
         else:
-            logger.warning( 'instance %s gave err (%s) "%s"', iid[0:8], type(err), err )
+            logger.warning( 'checkInstanceClocks instance %s gave status (%s) "%s"', iid[0:8], type(err), err )
         # change (or omit) this if wanting to allow more than one failure before marking it failed
         coll.update_one( {'_id': iid}, { "$set": { "state": "failed" } } )
 
+    checkables = [inst for inst in checkables if inst['instanceId'] not in errorsByIid]
+    logger.info( '%d instances checkable', len(checkables) )
+
 
     errorsByIid = checkGethProcesses( checkables, dataDirPath )
-    for iid, status in errorsByIid.items():
-        coll.update_one( {'_id': iid}, { "$inc": { "nFailures": 1 } } )
+    for iid, err in errorsByIid.items():
+        if 'status' in err:
+            status = err['status']
+            if isinstance( status, Exception ):
+                coll.update_one( {'_id': iid}, { "$inc": { "nExceptions": 1 } } )
+            else:
+                coll.update_one( {'_id': iid}, { "$inc": { "nFailures": 1 } } )
+        else:
+            logger.warning( 'checkGethProcesses for inst %s gave status (%s) "%s"', iid[0:8], type(err), err )
         # change (or omit) this if wanting to allow more than one failure before marking it failed
         coll.update_one( {'_id': iid}, { "$set": { "state": "failed" } } )
 
+    checkables = [inst for inst in checkables if inst['instanceId'] not in errorsByIid]
+    logger.info( '%d instances checkable', len(checkables) )
+
+    nodeDataDirPath = 'ether/%s' % args.configName
+    errorsByIid = retrieveLogs( checkables, dataDirPath, nodeDataDirPath )
+    checkLogs( checkables, dataDirPath )
+    processNodeLogFiles( dataDirPath )
+
     return
-    '''
-    resultsLogFilePath=dataDirPath+('/checkInstances_%s.jlog' % dateTimeTag)
-    workerCmd = 'stat --format=%%y ether/%s/geth.log' % args.configName
-    logger.info( 'calling tellInstances on %d instances', len(checkables))
-    stepStatuses = tellInstances.tellInstances( checkables, workerCmd,
-        resultsLogFilePath=resultsLogFilePath,
-        download=None, downloadDestDir=None, jsonOut=None, sshAgent=args.sshAgent,
-        timeLimit=checkerTimeLimit, upload=None, stopOnSigterm=True,
-        knownHostsOnly=False
-        )
-    '''
+
+def collectSigners( db, configName ):
+    anchorInstances = list( db['anchorInstances'].find() ) # fully iterates the cursor, getting all records
+    coll = db['checkedInstances']
+    wereChecked = list( coll.find({'state': 'checked'} ) )
+    for inst in wereChecked:
+        inst['instanceId'] = inst['_id']
+    instances = anchorInstances + wereChecked
+
+    # get the currently authorized signers (and true proposees)
+    signerInfos = ncsgeth.collectSignerInstances( instances, configName )
+    logger.debug( 'signerInfos: %s', signerInfos )
+    coll = db['allSigners']
+    logger.info( 'saving signers')
+    for signer in signerInfos:
+        signerId = signer['accountAddr']
+        iid = signer['instanceId']
+        auth = signer['auth']
+        logger.info( 'signer acct %s: iid %s, auth: %s', signerId, iid, auth )
+        coll.update_one( {'_id': iid}, { "$set": 
+            { "instanceId": iid, "accountAddr": signerId, "auth": auth } }, upsert=True
+            )
+
+def waitForAuth( victimAccount, shouldAuth, instances, configName, timeLimit ):
+    '''wait for the authorization of victimAccount to match shouldAuth; return success (true or false)'''
+    curSigners = ncsgeth.collectAuthSigners( instances, configName )
+    nowAuth = victimAccount in curSigners
+    logger.info( 'now authorized? %s', nowAuth )
+    deadline = time.time() + timeLimit
+    while nowAuth != shouldAuth:
+        if time.time() >= deadline:
+            logger.warning( 'took too long for auth/deauth to propagate')
+            break
+        curSigners = ncsgeth.collectAuthSigners( instances, configName )
+        nowAuth = victimAccount in curSigners
+        logger.info( 'now authorized? %s', nowAuth )
+        if nowAuth != shouldAuth:
+            time.sleep(1)
+    return nowAuth == shouldAuth
 
 
 if __name__ == "__main__":
@@ -576,7 +819,8 @@ if __name__ == "__main__":
     ap = argparse.ArgumentParser( description=__doc__,
         fromfile_prefix_chars='@', formatter_class=argparse.ArgumentDefaultsHelpFormatter )
     ap.add_argument( 'action', help='the action to perform', 
-        choices=['launch', 'import', 'check', 'terminateBad', 'terminateAll', 'reterminate']
+        choices=['launch', 'authorizeSigner', 'deauthorizeSigner', 'check', 'collectSigners', 'importAnchorNodes',
+            'terminateBad', 'terminateAll', 'reterminate']
         )
     ap.add_argument( '--authToken', help='the NCS authorization token to use (required for launch or terminate)' )
     ap.add_argument( '--count', type=int, help='the number of instances (for launch)' )
@@ -590,6 +834,7 @@ if __name__ == "__main__":
         default='{"dpr": ">=51", "ram:": ">=4000000000", "storage": ">=20000000000"}' )
     ap.add_argument( '--installerFileName', help='a script to run on the instances',
         default='netconfig/installAndStartGeth.sh' )
+    ap.add_argument( '--instanceId', help='id of instance (for authorize or deauthorize)' )
     ap.add_argument( '--uploads', help='glob for filenames to upload to workers', default='netconfig' )
     ap.add_argument( '--mongoHost', help='the host of mongodb server', default='localhost')
     ap.add_argument( '--mongoPort', help='the port of mongodb server', default=27017)
@@ -609,8 +854,70 @@ if __name__ == "__main__":
     db = mclient[dbName]
     if args.action == 'launch':
         launchEdgeNodes( dataDirPath, db, args )
+    elif args.action in ['authorizeSigner', 'deauthorizeSigner' ]:
+        victimIid = args.instanceId
+        configName = args.configName
+        shouldAuth = args.action == 'authorizeSigner'
+        logger.info( 'shouldAuth: %s', shouldAuth )
+
+        allSigners = list( db['allSigners'].find() )
+        logger.info( 'len(allSigners): %d', len(allSigners) )
+        signerIids = set( [signer['instanceId'] for signer in allSigners] )
+        logger.info( 'signerIids: %s', signerIids )
+
+        if (victimIid in signerIids) == shouldAuth:
+            logger.warning( 'instance %s auth status is already %s', victimIid, shouldAuth )
+            #sys.exit()
+
+        wereChecked = list( db['checkedInstances'].find( {'state': 'checked'}) ) # fully iterates the cursor, getting all records
+        logger.info( '%d instances were checked', len(wereChecked))
+        for inst in wereChecked:
+            inst['instanceId'] = inst['_id']
+        #checkedByIid = { inst['_id']: inst for inst in wereChecked }
+        anchorInstances = list( db['anchorInstances'].find() ) # fully iterates the cursor, getting all records
+
+        instances = anchorInstances + wereChecked
+        instancesByIid = {inst['instanceId']: inst for inst in instances }
+        victimInst = instancesByIid[ victimIid ]
+        if not victimInst:
+            logger.warning( 'instance %s not found', victimIid )
+            sys.exit()
+        logger.debug ('victimInst: %s', victimInst )
+
+        instanceAccountPairs = ncsgeth.collectPrimaryAccounts( [victimInst], configName )
+        if not instanceAccountPairs:
+            logger.warning( 'no account found for inst %s', victimIid )
+        victimAccount = instanceAccountPairs[0]['accountAddr']
+
+        authorizers = [inst for inst in instances if inst['instanceId'] in signerIids ]
+        logger.info( '%d authorizers', len(authorizers) )
+        if not authorizers:
+            logger.warning( 'no authorizers found')
+
+        authResults = ncsgeth.authorizeSigner( authorizers, configName, victimAccount, shouldAuth )
+        logger.info( 'authResults: %s', authResults )
+        anySucceeded = False
+        for result in authResults:
+            if 'returnCode' in result and result['returnCode'] == 0:
+                anySucceeded = True
+            else:
+                logger.warning( 'bad result from authorizeSigner: %s', result )
+        if anySucceeded:
+            itWorked = waitForAuth( victimAccount, shouldAuth, authorizers, configName, timeLimit=600 )
+            if itWorked:
+                db['allSigners'].update_one( {'_id': victimIid}, { "$set": 
+                    { "instanceId": victimIid, "accountAddr": victimAccount, "auth": shouldAuth }
+                     }, upsert=True
+                    )
+
+
     elif args.action == 'check':
         checkEdgeNodes( dataDirPath, db, args )
+    elif args.action == 'collectSigners':
+        collectSigners( db, args.configName )
+    elif args.action == 'importAnchorNodes':
+        anchorNodesFilePath = os.path.join( dataDirPath, 'anchorInstances.json' )
+        importAnchorNodes( anchorNodesFilePath, db, 'anchorInstances' )
     elif args.action == 'terminateBad':
         if not args.authToken:
             sys.exit( 'error: can not terminate because no authToken was passed')
